#include "driver.h"
#include "elvees_pci_user.h"
#include "elvees_pci_regs.h"

#include "trace.h"

#if defined(EVENT_TRACING)
#include "device_init.tmh"
#endif

#ifdef ALLOC_PRAGMA
#pragma alloc_text (PAGE, ElveesDeviceCreate)
#pragma alloc_text (PAGE, ElveesInitializeDma)
#pragma alloc_text (PAGE, ElveesEnablePciWindowAccess)
#pragma alloc_text (PAGE, ElveesEvtDevicePrepareHardware)
#pragma alloc_text (PAGE, ElveesEvtDeviceReleaseHardware)
#pragma alloc_text (PAGE, ElveesEvtD0Entry)
#endif


NTSTATUS
ElveesDeviceCreate(
    PWDFDEVICE_INIT DeviceInit
    )
/*++

Routine Description:

    Worker routine called to create a device and its software resources.

Arguments:

    DeviceInit - Pointer to an opaque init structure. Memory for this
                    structure will be freed by the framework when the WdfDeviceCreate
                    succeeds. So don't access the structure after that point.

Return Value:

    NTSTATUS

--*/
{
    WDF_OBJECT_ATTRIBUTES			deviceAttributes;
    WDF_PNPPOWER_EVENT_CALLBACKS    pnpPowerCallbacks;
	WDF_IO_QUEUE_CONFIG             ioQueueConfig;
	WDFQUEUE						ioctlQueue;
    WDFDEVICE						device;
    NTSTATUS						status;

	TraceEvents(TRACE_LEVEL_INFORMATION, DBG_INIT, "-->ElveesDeviceCreate\n");

    PAGED_CODE();

    WDF_PNPPOWER_EVENT_CALLBACKS_INIT(&pnpPowerCallbacks);

	pnpPowerCallbacks.EvtDevicePrepareHardware = ElveesEvtDevicePrepareHardware;
	pnpPowerCallbacks.EvtDeviceReleaseHardware = ElveesEvtDeviceReleaseHardware;
	pnpPowerCallbacks.EvtDeviceD0Entry = ElveesEvtD0Entry;

    //
    // Register the PnP and power callbacks. Power policy related callbacks will be registered
    // later in SotwareInit.
    //
    WdfDeviceInitSetPnpPowerEventCallbacks(DeviceInit, &pnpPowerCallbacks);

    WDF_OBJECT_ATTRIBUTES_INIT_CONTEXT_TYPE(&deviceAttributes, DEVICE_CONTEXT);

	//
	// By opting for SynchronizationScopeDevice, we tell the framework to
	// synchronize callbacks events of all the objects directly associated
	// with the device. In this driver, we will associate queues and
	// and DpcForIsr. By doing that we don't have to worrry about synchronizing
	// access to device-context by Io Events and DpcForIsr because they would
	// not concurrently ever. Framework will serialize them by using an
	// internal device-lock.
	//
	deviceAttributes.SynchronizationScope = WdfSynchronizationScopeDevice;

    status = WdfDeviceCreate(&DeviceInit, &deviceAttributes, &device);

	if (!NT_SUCCESS(status)) {
		TraceEvents(TRACE_LEVEL_ERROR, DBG_INIT, "WdfDeviceCreate failed with status %!STATUS!\n", status);
		return status;
	}

    //
    // Create a device interface so that application can find and talk
    // to us.
    //
    status = WdfDeviceCreateDeviceInterface(
        device,
        &GUID_DEVINTERFACE_ELVEES_PCI,
        NULL // ReferenceString
        );

	if (!NT_SUCCESS(status)) {
		TraceEvents(TRACE_LEVEL_ERROR, DBG_INIT, "WdfDeviceCreateDeviceInterface failed with status %!STATUS!\n", status);
		return status;
	}

	WDF_IO_QUEUE_CONFIG_INIT(
		&ioQueueConfig,
		WdfIoQueueDispatchSequential
		);

	ioQueueConfig.EvtIoDeviceControl = ElveesEvtIoDeviceControl;

	status = WdfIoQueueCreate(
		device,
		&ioQueueConfig,
		WDF_NO_OBJECT_ATTRIBUTES,
		&ioctlQueue
		);

	if (!NT_SUCCESS(status)) {
		TraceEvents(TRACE_LEVEL_ERROR, DBG_INIT, "WdfIoQueueCreate failed with status %!STATUS!\n", status);
		return status;
	}

	status = WdfDeviceConfigureRequestDispatching(
		device,
		ioctlQueue,
		WdfRequestTypeDeviceControl);

	if (!NT_SUCCESS(status)) {
		TraceEvents(TRACE_LEVEL_ERROR, DBG_INIT, "WdfDeviceConfigureRequestDispatching failed with status %!STATUS!\n", status);
		return status;
	}

	status = ElveesInitializeDma(device);
	if (!NT_SUCCESS(status)) {
		TraceEvents(TRACE_LEVEL_ERROR, DBG_INIT, "ElveesInitializeDma failed with status %!STATUS!\n", status);
		return status;
	}

	TraceEvents(TRACE_LEVEL_INFORMATION, DBG_INIT, "<--ElveesDeviceCreate\n");

    return status;
}

NTSTATUS
ElveesInitializeDma(
	IN WDFDEVICE device
)
{
	PDEVICE_CONTEXT			pContext;
	NTSTATUS				status;
	WDF_OBJECT_ATTRIBUTES	attributes;
	ULONG					dteCount;

	PAGED_CODE();

	pContext = GetContext(device);

	//
	// Set Maximum Transfer Length (which must be less than the SRAM size).
	//
	pContext->MaximumTransferLength = ELVEES_MAXIMUM_TRANSFER_LENGTH;

	TraceEvents(TRACE_LEVEL_INFORMATION, DBG_INIT,
		"MaximumTransferLength %d", pContext->MaximumTransferLength);

	//
	// Calculate the number of DMA_TRANSFER_ELEMENTS + 1 needed to
	// support the MaximumTransferLength.
	//
	dteCount = BYTES_TO_PAGES((ULONG)ROUND_TO_PAGES(
		pContext->MaximumTransferLength) + PAGE_SIZE);

	TraceEvents(TRACE_LEVEL_INFORMATION, DBG_INIT, "Number of DTEs %d", dteCount);

	//
	// Set the number of DMA_TRANSFER_ELEMENTs (DTE) to be available.
	//
	pContext->WriteTransferElements = dteCount;
	pContext->ReadTransferElements = dteCount;

	//
	// PLx PCI9656 DMA_TRANSFER_ELEMENTS must be 16-byte aligned
	//
	WdfDeviceSetAlignmentRequirement(device,
		FILE_LONG_ALIGNMENT);

	//
	// Create a new DMA Enabler instance.
	// Use Scatter/Gather, 64-bit Addresses, Duplex-type profile.
	//
	{
		WDF_DMA_ENABLER_CONFIG   dmaConfig;

		WDF_DMA_ENABLER_CONFIG_INIT(&dmaConfig,
			WdfDmaProfileScatterGather,
			pContext->MaximumTransferLength);

		TraceEvents(TRACE_LEVEL_INFORMATION, DBG_INIT,
			" - The DMA Profile is WdfDmaProfileScatterGather");

		status = WdfDmaEnablerCreate(device,
			&dmaConfig,
			WDF_NO_OBJECT_ATTRIBUTES,
			&pContext->DmaEnabler);

		if (!NT_SUCCESS(status)) {

			TraceEvents(TRACE_LEVEL_ERROR, DBG_INIT,
				"WdfDmaEnablerCreate failed: %!STATUS!", status);
			return status;
		}
	}

	//
	// Since we are using sequential queue and processing one request
	// at a time, we will create transaction objects upfront and reuse
	// them to do DMA transfer. Transactions objects are parented to
	// DMA enabler object by default. They will be deleted along with
	// along with the DMA enabler object. So need to delete them
	// explicitly.
	//
	WDF_OBJECT_ATTRIBUTES_INIT_CONTEXT_TYPE(&attributes, TRANSACTION_CONTEXT);

	status = WdfDmaTransactionCreate(pContext->DmaEnabler,
		&attributes,
		&pContext->ReadDmaTransaction);

	if (!NT_SUCCESS(status)) {
		TraceEvents(TRACE_LEVEL_ERROR, DBG_INIT,
			"WdfDmaTransactionCreate(read) failed: %!STATUS!", status);
		return status;
	}

	WDF_OBJECT_ATTRIBUTES_INIT_CONTEXT_TYPE(&attributes, TRANSACTION_CONTEXT);
	//
	// Create a new DmaTransaction.
	//
	status = WdfDmaTransactionCreate(pContext->DmaEnabler,
		&attributes,
		&pContext->WriteDmaTransaction);

	if (!NT_SUCCESS(status)) {
		TraceEvents(TRACE_LEVEL_ERROR, DBG_INIT,
			"WdfDmaTransactionCreate(write) failed: %!STATUS!", status);
		return status;
	}

	return status;
}

NTSTATUS
ElveesEnablePciWindowAccess(
	IN WDFDEVICE device
)
{
	BUS_INTERFACE_STANDARD			busInterface;
	NTSTATUS						status;
	DECLSPEC_ALIGN(MEMORY_ALLOCATION_ALIGNMENT) UCHAR buffer[ELVEES_PMSC_CONF_SIZE];
	PPCI_COMMON_CONFIG				pPciConfig = (PPCI_COMMON_CONFIG)buffer;
	ULONG							ulResult = 0;

	TraceEvents(TRACE_LEVEL_INFORMATION, DBG_INIT, "-->ElveesEnablePciWindowAccess\n");

	PAGED_CODE();

	status = WdfFdoQueryForInterface(device,
		&GUID_BUS_INTERFACE_STANDARD,
		(PINTERFACE)&busInterface,
		sizeof(BUS_INTERFACE_STANDARD),
		1, // Version
		NULL); //InterfaceSpecificData

	if (!NT_SUCCESS(status)) {
		TraceEvents(TRACE_LEVEL_ERROR, DBG_INIT, "WdfFdoQueryForInterface failed with status %!STATUS!\n", status);
		return status;
	}

	RtlZeroMemory(buffer, sizeof(buffer));
	ulResult = busInterface.GetBusData(
		busInterface.Context,
		PCI_WHICHSPACE_CONFIG, //READ
		buffer,
		0,
		ELVEES_PMSC_CONF_SIZE);

	if (ulResult != ELVEES_PMSC_CONF_SIZE) {
		TraceEvents(TRACE_LEVEL_ERROR, DBG_INIT, "GetBusData (ELVEES_PMSC_CONF_SIZE) failed with result %d\n", ulResult);
		return STATUS_INVALID_DEVICE_REQUEST;
	}

	pPciConfig->Command = 0x6;
	PUINT32 pCsrWin = (PUINT32)(buffer + 0x6C);
	*pCsrWin = (*pCsrWin & ~0x1E) | (0xF << 1);

	ulResult = busInterface.SetBusData(
		busInterface.Context,
		PCI_WHICHSPACE_CONFIG,
		buffer,
		0,
		ELVEES_PMSC_CONF_SIZE);

	if (ulResult == ELVEES_PMSC_CONF_SIZE) {
		status = STATUS_SUCCESS;
	}
	else {
		TraceEvents(TRACE_LEVEL_ERROR, DBG_INIT, "SetBusData failed with result %d, ELVEES_PMSC_CONF_SIZE = %d\n",
			ulResult, ELVEES_PMSC_CONF_SIZE);
		status = STATUS_UNSUCCESSFUL;
	}

	TraceEvents(TRACE_LEVEL_INFORMATION, DBG_INIT, "<--ElveesEnablePciWindowAccess\n");

	return status;
}

NTSTATUS
ElveesEvtDevicePrepareHardware(
WDFDEVICE      Device,
WDFCMRESLIST   Resources,
WDFCMRESLIST   ResourcesTranslated
)
/*++

Routine Description:

EvtDeviceStart event callback performs operations that are necessary
to make the driver's device operational. The framework calls the driver's
EvtDeviceStart callback when the PnP manager sends an IRP_MN_START_DEVICE
request to the driver stack.

Arguments:

Device - Handle to a framework device object.

Resources - Handle to a collection of framework resource objects.
This collection identifies the raw (bus-relative) hardware
resources that have been assigned to the device.

ResourcesTranslated - Handle to a collection of framework resource objects.
This collection identifies the translated (system-physical)
hardware resources that have been assigned to the device.
The resources appear from the CPU's point of view.
Use this list of resources to map I/O space and
device-accessible memory into virtual address space

Return Value:

WDF status code

--*/
{
	NTSTATUS     status = STATUS_SUCCESS;
	PDEVICE_CONTEXT	pContext;
	PCM_PARTIAL_RESOURCE_DESCRIPTOR descriptor;

	TraceEvents(TRACE_LEVEL_INFORMATION, DBG_INIT, "-->ElveesEvtDevicePrepareHardware\n");

	UNREFERENCED_PARAMETER(Resources);

	PAGED_CODE();

	pContext = GetContext(Device);

	for (ULONG i = 0; i<WdfCmResourceListGetCount(ResourcesTranslated); i++) {

		descriptor = WdfCmResourceListGetDescriptor(ResourcesTranslated, i);

		if (!descriptor) {
			TraceEvents(TRACE_LEVEL_ERROR, DBG_INIT, "WdfCmResourceListGetDescriptor failed, descriptor == 0\n");
			return STATUS_DEVICE_CONFIGURATION_ERROR;
		}

		switch (descriptor->Type) {
		case CmResourceTypeMemory:
			pContext->pciBaseAddrPhy = descriptor->u.Memory.Start;
			pContext->pciWindowLen = descriptor->u.Memory.Length;
			break;
		default:
			break;
		}
	}

	pContext->pciBaseAddrSys = MmMapIoSpace(pContext->pciBaseAddrPhy, 
		pContext->pciWindowLen,	MmNonCached);
	if (!pContext->pciBaseAddrSys) {
		TraceEvents(TRACE_LEVEL_ERROR, DBG_INIT, "MmMapIoSpace returned 0!\n");
		return STATUS_DEVICE_CONFIGURATION_ERROR;
	}

	TraceEvents(TRACE_LEVEL_INFORMATION, DBG_INIT, "<--ElveesEvtDevicePrepareHardware\n");

	return status;

}

NTSTATUS
ElveesEvtDeviceReleaseHardware(
IN  WDFDEVICE    Device,
IN  WDFCMRESLIST ResourcesTranslated
)
/*++

Routine Description:

EvtDeviceReleaseHardware is called by the framework whenever the PnP manager
is revoking ownership of our resources.  This may be in response to either
IRP_MN_STOP_DEVICE or IRP_MN_REMOVE_DEVICE.  The callback is made before
passing down the IRP to the lower driver.

In this callback, do anything necessary to free those resources.

Arguments:

Device - Handle to a framework device object.

ResourcesTranslated - Handle to a collection of framework resource objects.
This collection identifies the translated (system-physical)
hardware resources that have been assigned to the device.
The resources appear from the CPU's point of view.
Use this list of resources to map I/O space and
device-accessible memory into virtual address space

Return Value:

NTSTATUS - Failures will be logged, but not acted on.

--*/
{
	PDEVICE_CONTEXT	pContext;

	TraceEvents(TRACE_LEVEL_INFORMATION, DBG_INIT, "-->ElveesEvtDeviceReleaseHardware\n");

	UNREFERENCED_PARAMETER(ResourcesTranslated);

	PAGED_CODE();

	pContext = GetContext(Device);

	MmUnmapIoSpace(pContext->pciBaseAddrSys, pContext->pciWindowLen);

	TraceEvents(TRACE_LEVEL_INFORMATION, DBG_INIT, "<--ElveesEvtDeviceReleaseHardware\n");

	return STATUS_SUCCESS;
}

NTSTATUS
ElveesEvtD0Entry(
IN WDFDEVICE Device,
IN WDF_POWER_DEVICE_STATE PreviousState
)
{
	TraceEvents(TRACE_LEVEL_INFORMATION, DBG_INIT, "-->ElveesEvtD0Entry\n");

	UNREFERENCED_PARAMETER(Device);
	UNREFERENCED_PARAMETER(PreviousState);

	PAGED_CODE();

	ElveesEnablePciWindowAccess(Device);

	TraceEvents(TRACE_LEVEL_INFORMATION, DBG_INIT, "<--ElveesEvtD0Entry\n");

	return STATUS_SUCCESS;
}
