#include "driver.h"
#include "elvees_pci_user.h"
#include "elvees_pci_regs.h"

#include "trace.h"

#if defined(EVENT_TRACING)
#include "device_io.tmh"
#endif

VOID
ElveesEvtIoDeviceControl(
IN WDFQUEUE    Queue,
IN WDFREQUEST  Request,
IN size_t      OutputBufferLength,
IN size_t      InputBufferLength,
IN ULONG       IoControlCode
)
/*++

Routine Description:

This event is called when the framework receives IRP_MJ_DEVICE_CONTROL
requests from the system.

Arguments:

Queue - Handle to the framework queue object that is associated
with the I/O request.
Request - Handle to a framework request object.

OutputBufferLength - length of the request's output buffer,
if an output buffer is available.
InputBufferLength - length of the request's input buffer,
if an input buffer is available.

IoControlCode - the driver-defined or system-defined I/O control code
(IOCTL) that is associated with the request.
Return Value:

VOID

--*/
{
	NTSTATUS                status = STATUS_SUCCESS;
	PDEVICE_CONTEXT         pContext = NULL;
	WDFDEVICE               device;
	WDF_REQUEST_PARAMETERS  params;
	PUINT32					ioBuffer;
	PUINT32					rwBuffer;
	PUINT32					areaPointer;
	UINT32					areaOffset;
	size_t                  inputBufferLength;
	size_t					outputBufferLength;
	PTRANSACTION_CONTEXT	transContext;
	PMDL					mdl;
	PVOID					virtualAddress;
	ULONG					length;

	TraceEvents(TRACE_LEVEL_INFORMATION, DBG_IOCTLS, "-->ElveesEvtIoDeviceControl\n");

	device = WdfIoQueueGetDevice(Queue);
	pContext = GetContext(device);

	WDF_REQUEST_PARAMETERS_INIT(&params);

	WdfRequestGetParameters(
		Request,
		&params
		);

	switch (IoControlCode)
	{
	case IOCTL_ELVEES_READ_AREA:

		TraceEvents(TRACE_LEVEL_INFORMATION, DBG_IOCTLS, "IOCTL_ELVEES_READ_AREA\n");

		//      METHOD_BUFFERED, 
		//             .

		if (InputBufferLength != 4) {
			TraceEvents(TRACE_LEVEL_ERROR, DBG_IOCTLS, "Input buffer for ioctl request is too small!");
			WdfRequestComplete(Request, STATUS_INVALID_PARAMETER);
			return;
		}

		status = WdfRequestRetrieveInputBuffer(Request,
			4,
			&ioBuffer,
			&inputBufferLength);

		if (!NT_SUCCESS(status)) {
			TraceEvents(TRACE_LEVEL_ERROR, DBG_IOCTLS, "WdfRequestRetrieveInputBuffer failed 0x%x", status);
			WdfRequestComplete(Request, status);
			return;
		}

		areaOffset = *ioBuffer & ~3;

		status = WdfRequestRetrieveOutputBuffer(Request,
			OutputBufferLength,
			&ioBuffer,
			&outputBufferLength);

		if (!NT_SUCCESS(status)) {
			TraceEvents(TRACE_LEVEL_ERROR, DBG_IOCTLS, "WdfRequestRetrieveOutputBuffer failed 0x%x", status);
			WdfRequestComplete(Request, status);
			return;
		}

		outputBufferLength = (OutputBufferLength - 4 + 3) & ~3;

		if (areaOffset + outputBufferLength >= pContext->pciWindowLen) {
			TraceEvents(TRACE_LEVEL_ERROR, DBG_IOCTLS, "Output buffer for ioctl request is too small!");
			WdfRequestComplete(Request, STATUS_INVALID_PARAMETER);
			return;
		}

		//RtlCopyMemory(ioBuffer + 1, (PUCHAR)pContext->pciBaseAddrSys + areaOffset, outputBufferLength);

		areaPointer = (PUINT32)((PUCHAR)pContext->pciBaseAddrSys + areaOffset);
		rwBuffer = ioBuffer + 1;
		while (outputBufferLength) {
			*rwBuffer = *areaPointer;
			rwBuffer++;
			areaPointer++;
			outputBufferLength -= 4;
		}

		WdfRequestCompleteWithInformation(Request, status, (OutputBufferLength + 3) & ~3);

		break;

	case IOCTL_ELVEES_WRITE_AREA:

		TraceEvents(TRACE_LEVEL_INFORMATION, DBG_IOCTLS, "IOCTL_ELVEES_WRITE_AREA\n");

		//      METHOD_BUFFERED, 
		//             .

		if (InputBufferLength < 8) {
			WdfRequestComplete(Request, STATUS_INVALID_PARAMETER);
			return;
		}

		status = WdfRequestRetrieveInputBuffer(Request,
			InputBufferLength,
			&ioBuffer,
			&inputBufferLength);

		if (!NT_SUCCESS(status)) {
			TraceEvents(TRACE_LEVEL_ERROR, DBG_IOCTLS, "WdfRequestRetrieveInputBuffer failed 0x%x", status);
			WdfRequestComplete(Request, status);
			return;
		}

		areaOffset = *ioBuffer;

		if (areaOffset + OutputBufferLength >= pContext->pciWindowLen) {
			TraceEvents(TRACE_LEVEL_ERROR, DBG_IOCTLS, "Output buffer for ioctl request is too small!");
			WdfRequestComplete(Request, STATUS_INVALID_PARAMETER);
			return;
		}

		//RtlCopyMemory((PUCHAR)pContext->pciBaseAddrSys + areaOffset, ioBuffer + 1, InputBufferLength - 4);

		inputBufferLength = (InputBufferLength - 4 + 3) & ~3;
		areaPointer = (PUINT32)((PUCHAR)pContext->pciBaseAddrSys + areaOffset);
		rwBuffer = ioBuffer + 1;
		while (inputBufferLength) {
			*areaPointer = *rwBuffer;
			rwBuffer++;
			areaPointer++;
			inputBufferLength -= 4;
		}

		WdfRequestComplete(Request, status);

		break;

	case IOCTL_ELVEES_READ_AREA_DMA:

		TraceEvents(TRACE_LEVEL_INFORMATION, DBG_IOCTLS, "IOCTL_ELVEES_READ_AREA_DMA\n");

		if (InputBufferLength != 4) {
			TraceEvents(TRACE_LEVEL_ERROR, DBG_IOCTLS, "Input buffer for ioctl request is too small!");
			WdfRequestComplete(Request, STATUS_INVALID_PARAMETER);
			return;
		}

		status = WdfRequestRetrieveInputBuffer(Request,
			4,
			&ioBuffer,
			&inputBufferLength);

		if (!NT_SUCCESS(status)) {
			TraceEvents(TRACE_LEVEL_ERROR, DBG_IOCTLS, "WdfRequestRetrieveInputBuffer failed 0x%x", status);
			WdfRequestComplete(Request, status);
			return;
		}

		areaOffset = *ioBuffer & ~3;

		//
		// Initialize this new DmaTransaction with direct parameters.
		//
		status = WdfRequestRetrieveOutputWdmMdl(Request, &mdl);
		if (!NT_SUCCESS(status)) {
			TraceEvents(TRACE_LEVEL_ERROR, DBG_IOCTLS,
				"WdfRequestRetrieveInputWdmMdl failed: %!STATUS!", status);
			WdfRequestComplete(Request, status);
			break;
		}

		virtualAddress = MmGetMdlVirtualAddress(mdl);
		length = MmGetMdlByteCount(mdl);

		TraceEvents(TRACE_LEVEL_INFORMATION, DBG_IOCTLS,
			"WdfRequestRetrieveInputWdmMdl: virtual address %p, length %d",
			virtualAddress, length);

		_Analysis_assume_(length > 0);
		status = WdfDmaTransactionInitialize(
			pContext->ReadDmaTransaction,
			ElveesEvtProgramDma,
			WdfDmaDirectionReadFromDevice,
			mdl,
			(UINT32 *)virtualAddress + 1,
			length - 4);

		if (!NT_SUCCESS(status)) {
			TraceEvents(TRACE_LEVEL_ERROR, DBG_IOCTLS,
				"WdfDmaTransactionInitialize failed: %!STATUS!", status);
			WdfDmaTransactionRelease(pContext->ReadDmaTransaction);
			WdfRequestComplete(Request, status);
			break;
		}

		//
		// Retreive this DmaTransaction's context ptr (aka TRANSACTION_CONTEXT)
		// and fill it in with info.
		//
		transContext = GetTransactionContext(pContext->ReadDmaTransaction);
		transContext->Request = Request;
		transContext->ElveesAddress = areaOffset;

		//
		// Execute this DmaTransaction transaction.
		//
		status = WdfDmaTransactionExecute(pContext->ReadDmaTransaction,
			WDF_NO_CONTEXT);

		if (!NT_SUCCESS(status)) {
			//
			// Couldn't execute this DmaTransaction, so fail Request.
			//
			TraceEvents(TRACE_LEVEL_ERROR, DBG_IOCTLS,
				"WdfDmaTransactionExecute failed: %!STATUS!", status);
			WdfDmaTransactionRelease(pContext->ReadDmaTransaction);
			WdfRequestComplete(Request, status);
			break;
		}

		break;

	case IOCTL_ELVEES_WRITE_AREA_DMA:

		TraceEvents(TRACE_LEVEL_INFORMATION, DBG_IOCTLS, "IOCTL_ELVEES_WRITE_AREA_DMA\n");

		if (InputBufferLength < 8) {
			WdfRequestComplete(Request, STATUS_INVALID_PARAMETER);
			return;
		}

		status = WdfRequestRetrieveInputBuffer(Request,
			4,
			&ioBuffer,
			&inputBufferLength);

		if (!NT_SUCCESS(status)) {
			TraceEvents(TRACE_LEVEL_ERROR, DBG_IOCTLS, "WdfRequestRetrieveInputBuffer failed 0x%x", status);
			WdfRequestComplete(Request, status);
			return;
		}

		areaOffset = *ioBuffer & ~3;

		//
		// Initialize this new DmaTransaction with direct parameters.
		//
		status = WdfRequestRetrieveInputWdmMdl(Request, &mdl);
		if (!NT_SUCCESS(status)) {
			TraceEvents(TRACE_LEVEL_ERROR, DBG_IOCTLS,
				"WdfRequestRetrieveInputWdmMdl failed: %!STATUS!", status);
			WdfRequestComplete(Request, status);
			break;
		}

		virtualAddress = MmGetMdlVirtualAddress(mdl);
		length = MmGetMdlByteCount(mdl);

		TraceEvents(TRACE_LEVEL_INFORMATION, DBG_IOCTLS,
			"WdfRequestRetrieveInputWdmMdl: virtual address %p, length %d",
			virtualAddress, length);

		_Analysis_assume_(length > 0);
		status = WdfDmaTransactionInitialize(
			pContext->WriteDmaTransaction,
			ElveesEvtProgramDma,
			WdfDmaDirectionWriteToDevice,
			mdl,
			(UINT32 *)virtualAddress + 1,
			length - 4);

		if (!NT_SUCCESS(status)) {
			TraceEvents(TRACE_LEVEL_ERROR, DBG_IOCTLS,
				"WdfDmaTransactionInitialize failed: %!STATUS!", status);
			WdfDmaTransactionRelease(pContext->WriteDmaTransaction);
			WdfRequestComplete(Request, status);
			break;
		}

		//
		// Retreive this DmaTransaction's context ptr (aka TRANSACTION_CONTEXT)
		// and fill it in with info.
		//
		transContext = GetTransactionContext(pContext->WriteDmaTransaction);
		transContext->Request = Request;
		transContext->ElveesAddress = areaOffset;

		//
		// Execute this DmaTransaction transaction.
		//
		status = WdfDmaTransactionExecute(pContext->WriteDmaTransaction,
			WDF_NO_CONTEXT);

		if (!NT_SUCCESS(status)) {
			//
			// Couldn't execute this DmaTransaction, so fail Request.
			//
			TraceEvents(TRACE_LEVEL_ERROR, DBG_IOCTLS,
				"WdfDmaTransactionExecute failed: %!STATUS!", status);
			WdfDmaTransactionRelease(pContext->WriteDmaTransaction);
			WdfRequestComplete(Request, status);
			break;
		}

		break;

	default:
		TraceEvents(TRACE_LEVEL_ERROR, DBG_IOCTLS, "Invalid IOCTL!");
		WdfRequestComplete(Request, STATUS_INVALID_DEVICE_REQUEST);
		break;
	}

	TraceEvents(TRACE_LEVEL_INFORMATION, DBG_IOCTLS, "<--ElveesEvtIoDeviceControl\n");

	return;
}

BOOLEAN
ElveesEvtProgramDma(
IN  WDFDMATRANSACTION       Transaction,
IN  WDFDEVICE               Device,
IN  PVOID                   Context,
IN  WDF_DMA_DIRECTION       Direction,
IN  PSCATTER_GATHER_LIST    SgList
)
{
	UNREFERENCED_PARAMETER(Context);

	NTSTATUS                 status = STATUS_SUCCESS;
	PDEVICE_CONTEXT          pContext;
	PTRANSACTION_CONTEXT     transContext;
	UINT32                   mcbAddr;
	ULONG                    i;

	TraceEvents(TRACE_LEVEL_INFORMATION, DBG_IOCTLS,
		"--> ElveesEvtProgramWriteDma");

	pContext = GetContext(Device);

	transContext = GetTransactionContext(Transaction);

	mcbAddr = transContext->ElveesAddress;

	for (i = 0; i < SgList->NumberOfElements; i++) {

		while (*(PUINT32)((PUCHAR)pContext->pciBaseAddrSys + PCI_STATUS_MASTER) & PCI_STATUS_MASTER_RUN);

		*(PUINT32)((PUCHAR)pContext->pciBaseAddrSys + PCI_IR_MASTER) = mcbAddr;
		*(PUINT32)((PUCHAR)pContext->pciBaseAddrSys + PCI_AR_PCI) = SgList->Elements[i].Address.LowPart;

		TraceEvents(TRACE_LEVEL_INFORMATION, DBG_IOCTLS,
			"WdfRequestRetrieveInputWdmMdl: virtual address %08X, mcbAddr %08X",
			SgList->Elements[i].Address.LowPart, mcbAddr);

		*(PUINT32)((PUCHAR)pContext->pciBaseAddrSys + PCI_CSR_MASTER) = PCI_CSR_MASTER_RUN |
			((Direction == WdfDmaDirectionWriteToDevice) ? PCI_CSR_MASTER_MEMREAD : PCI_CSR_MASTER_MEMWRITE) |
			PCI_CSR_MASTER_WC((SgList->Elements[i].Length >> 2) - 1) | PCI_CSR_MASTER_WNM(0xF);

		while (*(PUINT32)((PUCHAR)pContext->pciBaseAddrSys + PCI_STATUS_MASTER) & PCI_STATUS_MASTER_RUN);

		mcbAddr += SgList->Elements[i].Length;

	}

	WdfDmaTransactionRelease(Transaction);
	WdfRequestCompleteWithInformation(transContext->Request, status, mcbAddr - transContext->ElveesAddress);

	TraceEvents(TRACE_LEVEL_INFORMATION, DBG_IOCTLS,
		"<-- ElveesEvtProgramWriteDma");

	return TRUE;
}
