#ifdef __linux

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <stdint.h>
#include "libpe.h"

static int uio_number;
static int uio_fd;

static uint8_t *mem_addr;
static int mem_size;

int libpe_init(void)
{
    char filename[256];
    FILE *fp;
    char data[16];

    uio_number = 0;
    while (1) {
        sprintf(filename, "/sys/class/uio/uio%d/name", uio_number);
        fp = fopen(filename, "r");
        if (!fp) {
            perror("Error opening file uio/name");
            return 0;
        }
        if (fread(data, 3, 1, fp) != 1) {
            perror("Error reading from file uio/name");
            fclose(fp);
            return 0;
        }
        fclose(fp);
        if (strncmp(data, "MCB", 3) == 0)
            break;
    }

    sprintf(filename, "/sys/class/uio/uio%d/maps/map0/name", uio_number);
    fp = fopen(filename, "r");
    if (!fp) {
        perror("Error opening file uio/map/name");
        return 0;
    }
    if (fscanf(fp, "%s", data) != 1) {
        perror("Error reading from file uio/map/name");
        fclose(fp);
        return 0;
    }
    if (strcmp(data, "mcb_window") != 0) {
        fprintf(stderr, "Unknown area name: %s\n", data);
        return 0;
    }
    fclose(fp);

    sprintf(filename, "/sys/class/uio/uio%d/maps/map0/size", uio_number);
    fp = fopen(filename, "r");
    if (!fp) {
        perror("Error opening file uio/map/size");
        return 0;
    }
    if (fscanf(fp, "0x%X", &mem_size) != 1) {
        perror("Error reading from file uio/map/size");
        fclose(fp);
        return 0;
    }
    fclose(fp);

    sprintf(filename, "/dev/uio%d", uio_number);

    uio_fd = open(filename, O_RDWR);
    if (uio_fd < 0) {
        perror("Error opening file /dev/uio_");
        return 0;
    }

    mem_addr = mmap(NULL, mem_size,
        PROT_READ | PROT_WRITE, MAP_SHARED, uio_fd, 0);

    if (mem_addr == MAP_FAILED) {
        perror("Error mapping area");
        mem_addr = NULL;
        return 0;
    }

    return 1;
}

void libpe_close()
{
    munmap(mem_addr, mem_size);

    close(uio_fd);
}

int libpe_read(void *buffer, size_t size)
{
    uint32_t *pbuf = buffer;
    //uint32_t *pmem = (uint32_t *)(mem_addr + *pbuf);
    size &= ~0x3;
    //size_t sz = size - 4;

    /*
    pbuf++;
    while (sz) {
        *pbuf++ = *pmem++;
        sz -= 4;
    }
    */
    memcpy(pbuf + 1, mem_addr + *pbuf, size - 4);
    return size - 4;
}

int libpe_write(void *buffer, size_t size)
{
    uint32_t *pbuf = buffer;
    //uint32_t *pmem = (uint32_t *)(mem_addr + *pbuf);
    size &= ~0x3;
    //size_t sz = size - 4;

    /*
    pbuf++;
    while (sz) {
        *pmem++ = *pbuf++;
        sz -= 4;
    }
    */
    memcpy(mem_addr + *pbuf, pbuf + 1, size - 4);
    return size - 4;
}

void *libpe_raw_pointer_linux(void)
{
    return mem_addr;
}

#else

#define INITGUID
#include <windows.h>
#include <strsafe.h>
#include <setupapi.h>
#include <stdio.h>
#include <stdlib.h>
#include <devguid.h>
#include "elvees_pci_public.h"
#include "elvees_pci_user.h"

#define MAX_DEVPATH_LENGTH                       256

// Граница размера пакета при чтении, до которой используется
// обращение к плате без DMA, выше - с использованием DMA.
#define PIO_DMA_EDGE_RD                          16

// Граница размера пакета при записи, до которой используется
// обращение к плате без DMA, выше - с использованием DMA.
#define PIO_DMA_EDGE_WR                          160

static HANDLE hDevice;


BOOL GetDevicePath(LPGUID InterfaceGuid, PWCHAR DevicePath, size_t BufLen);


int libpe_init(void)
{
    WCHAR G_DevicePath[MAX_DEVPATH_LENGTH];
    hDevice = INVALID_HANDLE_VALUE;

    if ( !GetDevicePath(
            (LPGUID) &GUID_DEVINTERFACE_ELVEES_PCI,
            G_DevicePath,
            sizeof(G_DevicePath)/sizeof(G_DevicePath[0])) )
    {
        return 0;
    }

    hDevice = CreateFile(G_DevicePath,
                         GENERIC_READ|GENERIC_WRITE,
                         FILE_SHARE_READ | FILE_SHARE_WRITE,
                         NULL,
                         OPEN_EXISTING,
                         0,
                         NULL );

    if (hDevice == INVALID_HANDLE_VALUE) {
        return 0;
    }

    return 1;
}

void libpe_close(void)
{
    if (hDevice != INVALID_HANDLE_VALUE) {
        CloseHandle(hDevice);
    }
}

int libpe_read(void *buffer, size_t size)
{
    DWORD       BytesReturned;
    if (size <= PIO_DMA_EDGE_RD + 4)
        return DeviceIoControl(hDevice,	(DWORD)IOCTL_ELVEES_READ_AREA,
            buffer, 4, buffer, size, &BytesReturned, NULL);
    else
        return DeviceIoControl(hDevice,	(DWORD)IOCTL_ELVEES_READ_AREA_DMA,
            buffer, 4, buffer, size, &BytesReturned, NULL);
}

int libpe_write(void *buffer, size_t size)
{
    DWORD       BytesReturned;

    if (size <= PIO_DMA_EDGE_WR + 4)
        return DeviceIoControl(hDevice,	(DWORD)IOCTL_ELVEES_WRITE_AREA,
            buffer, size, 0, 0, &BytesReturned, NULL);
    else
        return DeviceIoControl(hDevice,	(DWORD)IOCTL_ELVEES_WRITE_AREA_DMA,
            buffer, size, 0, 0, &BytesReturned, NULL);
}

BOOL
GetDevicePath(
    LPGUID InterfaceGuid,
    PWCHAR DevicePath,
    size_t BufLen
    )
{
    HDEVINFO HardwareDeviceInfo;
    SP_DEVICE_INTERFACE_DATA DeviceInterfaceData;
    PSP_DEVICE_INTERFACE_DETAIL_DATA DeviceInterfaceDetailData = NULL;
    ULONG Length, RequiredLength = 0;
    BOOL bResult;
    HRESULT hr;

    HardwareDeviceInfo = SetupDiGetClassDevs(
                             InterfaceGuid,
                             NULL,
                             NULL,
                             (DIGCF_PRESENT | DIGCF_DEVICEINTERFACE));

    if (HardwareDeviceInfo == INVALID_HANDLE_VALUE) {
        return FALSE;
    }

    DeviceInterfaceData.cbSize = sizeof(SP_DEVICE_INTERFACE_DATA);

    bResult = SetupDiEnumDeviceInterfaces(HardwareDeviceInfo,
                                              0,
                                              InterfaceGuid,
                                              0,
                                              &DeviceInterfaceData);

    if (bResult == FALSE) {
        SetupDiDestroyDeviceInfoList(HardwareDeviceInfo);
        return FALSE;
    }

    SetupDiGetDeviceInterfaceDetail(
        HardwareDeviceInfo,
        &DeviceInterfaceData,
        NULL,
        0,
        &RequiredLength,
        NULL
        );

    DeviceInterfaceDetailData = (PSP_DEVICE_INTERFACE_DETAIL_DATA)LocalAlloc(LMEM_FIXED, RequiredLength);

    if (DeviceInterfaceDetailData == NULL) {
        SetupDiDestroyDeviceInfoList(HardwareDeviceInfo);
        return FALSE;
    }

    DeviceInterfaceDetailData->cbSize = sizeof(SP_DEVICE_INTERFACE_DETAIL_DATA);

    Length = RequiredLength;

    bResult = SetupDiGetDeviceInterfaceDetail(
                  HardwareDeviceInfo,
                  &DeviceInterfaceData,
                  DeviceInterfaceDetailData,
                  Length,
                  &RequiredLength,
                  NULL);

    if (bResult == FALSE) {
        SetupDiDestroyDeviceInfoList(HardwareDeviceInfo);
        LocalFree(DeviceInterfaceDetailData);
        return FALSE;
    }

    hr = StringCchCopy(DevicePath,
                    BufLen,
                    DeviceInterfaceDetailData->DevicePath) ;

    SetupDiDestroyDeviceInfoList(HardwareDeviceInfo);
    LocalFree(DeviceInterfaceDetailData);

    return ( !FAILED(hr) );
}

void *libpe_raw_pointer_linux(void)
{
    return NULL;
}


#endif
