#include <QCoreApplication>
#include <QDebug>
#include <QTime>
#include <stdint.h>
#include <string.h>
#include "libpe.h"

#ifdef __linux
#include <unistd.h>
#define Sleep(x) usleep(x)
#endif

#define MCB_VEND_ID 0x680C2001

#define MAX_BLOCK_SIZE 65536

uint8_t buf[MAX_BLOCK_SIZE + 4];
unsigned block_size = 4096;
unsigned nb_run = 1;

void mcb_close_and_exit()
{
    qDebug() << "\nExiting... Press ENTER...";

    libpe_close();

    getchar();
    exit(1);
}

void fatal_error()
{
#ifdef __linux
    qDebug() << "\nDRIVER INPUT/OUTPUT ERROR!";
#else
    qDebug() << "\nDRIVER INPUT/OUTPUT ERROR:" << GetLastError() << "!";
#endif
    mcb_close_and_exit();
}

uint32_t mcb_read_reg (uint32_t addr)
{
    uint32_t buf[2];
    buf[0] = addr;
    if (!libpe_read(buf, sizeof(buf)))
        fatal_error();

    return buf[1];
}

void mcb_write_reg(uint32_t addr, uint32_t value)
{
    uint32_t buf[2];

    buf[0] = addr;
    buf[1] = value;

    if (!libpe_write(buf, sizeof(buf)))
        fatal_error();
}

void test()
{
    qDebug("Reading DPRAM using %d byte blocks", block_size);
    qDebug("Number of run: %u", nb_run);

    unsigned overall = 0;
    QTime timer;

    timer.start();

    for (unsigned run = 0; run < nb_run; ++run) {
        overall = 0;
        while (overall + block_size < MCB_DPRAM_SIZE * 4) {
            uint32_t *paddr = (uint32_t *) buf;
            *paddr = MCB_DPRAM_BASE(0) + overall;
            libpe_read(buf, block_size + 4);
            overall += block_size;
        }
    }

    int elapsed_ms = timer.elapsed();
    qDebug("Reading DPRAM finished, %d ms elapsed, speed rate %.02f kbytes/sec",
           elapsed_ms, (float)overall * nb_run / elapsed_ms);
}

void usage()
{
    qDebug() << "Usage:";
    qDebug() << "    test-read-speed [block size] [number of run]";
    qDebug() << "Block size is set in bytes. Default value: 4096 bytes.";
    exit(2);
}

int main(int argc, char *argv[])
{
    QCoreApplication a(argc, argv);

    qDebug() << "\n=== MCB-03PEM-PCI READ SPEED TEST ===\n";

    if (a.arguments().count() > 1)
        block_size = a.arguments()[1].toUInt();
    if (a.arguments().count() > 2)
        nb_run = a.arguments()[2].toUInt();
    if (a.arguments().count() > 3)
        usage();

    if (block_size == 0 || block_size > MAX_BLOCK_SIZE) {
        qDebug() << "Unsupported block size!";
        usage();
    }

    if (!libpe_init()) {
        perror("Failed to initialize libmcb");
        exit(1);
    }

    qDebug("VENDOR ID/DEVICE ID: %08X\n",   mcb_read_reg(MCB_PCI_DEVICE_VENDOR_ID));
    qDebug("SWIC HW VERSION:     %08X\n\n", mcb_read_reg(MCB_SWIC_HW_VER(0)));

    test();

    mcb_close_and_exit();

    return 0;
}
