#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include "libpe.h"

#ifdef __linux
#include <unistd.h>
#define Sleep(x) usleep(x)
#endif

#define MCB_VEND_ID 0x680C2001

void mcb_close_and_exit()
{
    printf("\nExiting... Press ENTER...\n");

    libpe_close();

    getchar();
    exit(1);
}

void fatal_error()
{
#ifdef __linux
    printf("\nDRIVER INPUT/OUTPUT ERROR!\n");
#else
    printf("\nDRIVER INPUT/OUTPUT ERROR: %ld!\n", GetLastError());
#endif
    mcb_close_and_exit();
}

uint32_t mcb_read_reg (uint32_t addr)
{
    uint32_t buf[2];
    buf[0] = addr;
    if (!libpe_read(buf, sizeof(buf)))
        fatal_error();

    return buf[1];
}

void mcb_write_reg(uint32_t addr, uint32_t value)
{
    uint32_t buf[2];

    buf[0] = addr;
    buf[1] = value;

    if (!libpe_write(buf, sizeof(buf)))
        fatal_error();
}

#define	MAX_REP	1000

#define	DPRAM_HALF      0x20000
#define NB_WORDS        32766

int             nRep;
int             nErrors;
int             spd;
int             tx_chan;
int             rx_chan;

uint32_t        qstrConnectMask;
uint32_t        qstrErrMask;
uint32_t        dmaFinMask;


void usage()
{
    printf("Usage:\n");
    printf("    test-swic <number of test run> <speed in MHz> <tx channel> <rx channel>\n");
    exit(2);
}

void parse_args(int argc, char **argv)
{
    if (argc < 5)
        usage();

    nRep = strtoul(argv[1], NULL, 10);
    if (nRep == 0)
        usage();

    spd = strtoul(argv[2], NULL, 16);

    tx_chan = strtoul(argv[3], NULL, 10);
    if (tx_chan > 3)
        usage();

    rx_chan = strtoul(argv[4], NULL, 10);
    if (rx_chan > 3 || rx_chan == tx_chan)
        usage();

}

void print_status()
{
    printf("TX CHANNEL STATUS:  %08X\n", mcb_read_reg(MCB_SWIC_STATUS(tx_chan)));
    printf("RX CHANNEL STATUS:  %08X\n", mcb_read_reg(MCB_SWIC_STATUS(rx_chan)));
    printf("RX DESC DMA CSR:    %08X\n", mcb_read_reg(MCB_SWIC_RX_DESC_CSR(rx_chan)));
    printf("RX DATA DMA CSR:    %08X\n", mcb_read_reg(MCB_SWIC_RX_DATA_CSR(rx_chan)));
    printf("TX DESC DMA CSR:    %08X\n", mcb_read_reg(MCB_SWIC_TX_DESC_CSR(tx_chan)));
    printf("TX DATA DMA CSR:    %08X\n", mcb_read_reg(MCB_SWIC_TX_DATA_CSR(tx_chan)));
}

void clean_dma(int port)
{
    uint32_t mode;

    mode = mcb_read_reg (MCB_SWIC_MODE_CR (port));
    mode &= ~ (MC_SWIC_AutoStart | MC_SWIC_LinkStart);
    mode |= MC_SWIC_LinkDisabled;
    mcb_write_reg (MCB_SWIC_MODE_CR (port), mode);

    mcb_write_reg (MCB_SWIC_RX_DESC_RUN (port), 0);
    mcb_write_reg (MCB_SWIC_RX_DATA_RUN (port), 0);
    mcb_write_reg (MCB_SWIC_TX_DESC_RUN (port), 0);
    mcb_write_reg (MCB_SWIC_TX_DATA_RUN (port), 0);

    mode |= MC_SWIC_SWCORE_RST;
    //mode |= MC_SWIC_DS_RST;
    mcb_write_reg (MCB_SWIC_MODE_CR (port), mode);

    while ((mcb_read_reg (MCB_SWIC_RX_DESC_RUN (port)) |
            mcb_read_reg (MCB_SWIC_RX_DATA_RUN (port)) |
            mcb_read_reg (MCB_SWIC_TX_DESC_RUN (port)) |
            mcb_read_reg (MCB_SWIC_TX_DATA_RUN (port))) & 1);

    mcb_read_reg (MCB_SWIC_RX_DESC_CSR (port));
    mcb_read_reg (MCB_SWIC_RX_DATA_CSR (port));
    mcb_read_reg (MCB_SWIC_TX_DESC_CSR (port));
    mcb_read_reg (MCB_SWIC_TX_DATA_CSR (port));

    mode &= ~MC_SWIC_SWCORE_RST;
    //mode &= ~MC_SWIC_DS_RST;
    mcb_write_reg (MCB_SWIC_MODE_CR (port), mode);

    mcb_write_reg (MCB_SWIC_RX_DESC_CP (port), 0);
    mcb_write_reg (MCB_SWIC_RX_DATA_CP (port), 0);
    mcb_write_reg (MCB_SWIC_TX_DESC_CP (port), 0);
    mcb_write_reg (MCB_SWIC_TX_DATA_CP (port), 0);

    mcb_write_reg (MCB_SWIC_RX_DESC_IR (port), 0);
    mcb_write_reg (MCB_SWIC_RX_DATA_IR (port), 0);
    mcb_write_reg (MCB_SWIC_TX_DESC_IR (port), 0);
    mcb_write_reg (MCB_SWIC_TX_DATA_IR (port), 0);

    mcb_write_reg (MCB_SWIC_RX_DESC_CSR (port), 0);
    mcb_write_reg (MCB_SWIC_RX_DATA_CSR (port), 0);
    mcb_write_reg (MCB_SWIC_TX_DESC_CSR (port), 0);
    mcb_write_reg (MCB_SWIC_TX_DATA_CSR (port), 0);
}

#define SWIC_RESET              0x00000079
#define SWIC_START              0x00000101

int init_hw()
{
    clean_dma(tx_chan);
    clean_dma(rx_chan);

    mcb_write_reg(MCB_SWIC_MODE_CR(tx_chan), SWIC_RESET);
    mcb_write_reg(MCB_SWIC_MODE_CR(rx_chan), SWIC_RESET);
    mcb_write_reg(MCB_SWIC_MODE_CR(tx_chan), SWIC_START);
    mcb_write_reg(MCB_SWIC_MODE_CR(rx_chan), SWIC_START);
    mcb_write_reg(MCB_SWIC_TX_SPEED(tx_chan), MC_SWIC_PLL_TX_EN | MC_SWIC_LVDS_EN | MC_SWIC_TX_SPEED_PRM(2));
    mcb_write_reg(MCB_SWIC_TX_SPEED(rx_chan), MC_SWIC_PLL_TX_EN | MC_SWIC_LVDS_EN | MC_SWIC_TX_SPEED_PRM(2));
    mcb_write_reg(MCB_SWIC_MODE_CR(tx_chan), MC_SWIC_LinkStart | MC_SWIC_AutoStart | MC_SWIC_WORK_TYPE);
    mcb_write_reg(MCB_SWIC_MODE_CR(rx_chan), MC_SWIC_LinkStart | MC_SWIC_AutoStart | MC_SWIC_WORK_TYPE);

    int timeout = 10000;
    uint32_t src_status;
    uint32_t dst_status;
    /* wait for connect */
    while (timeout--) {
        src_status = mcb_read_reg(MCB_SWIC_STATUS(tx_chan));
        dst_status = mcb_read_reg(MCB_SWIC_STATUS(rx_chan));
        if (((src_status & 0x10F0) == 0x10A0) && ((dst_status & 0x10F0) == 0x10A0)) {
            if (src_status & 0xF)
                mcb_write_reg(MCB_SWIC_STATUS(tx_chan), 0xF);
            if (dst_status & 0xF)
                mcb_write_reg(MCB_SWIC_STATUS(rx_chan), 0xF);
            return 1;
        }

        Sleep(10);
    }

    return 0;

    //mcb_close_and_exit();
}

void test()
{
    qstrConnectMask = (1 << (3 * tx_chan)) | (1 << (3 * rx_chan));
    qstrErrMask = (2 << (3 * rx_chan));
    dmaFinMask = (0x3000 << (4 * rx_chan));

    printf("qstrConnectMask = %X\n", qstrConnectMask);
    printf("qstrErrMask = %X\n", qstrErrMask);
    printf("dmaFinMask = %X\n", dmaFinMask);

    nErrors	= 0;

    int step;
    for (step = 0; step < nRep; step++) {

        printf("SWICs initialization...\n");
        int try_nb = 0;
        while (!init_hw()) {
            try_nb++;
            if (try_nb > 10) {
                printf("\n--> Unable to connect SWICs!\n\n");
                nErrors++;
                continue;
            }
        }
        printf("\n--> SWICs are connected!\n\n");
        print_status();

        // set tx speed
        mcb_write_reg(MCB_SWIC_TX_SPEED(tx_chan),
                      MC_SWIC_PLL_TX_EN | MC_SWIC_LVDS_EN | MC_SWIC_TX_SPEED_PRM(spd/5));

        // timeout 100 ms, wait for events
        int timeout = 100;
        uint32_t qstr;
        while (timeout) {
            Sleep(10);
            timeout -= 10;

            qstr = mcb_read_reg(MCB_PCI_QSTR_PCI);
            if (qstr & qstrErrMask) {
                printf("\n--> SWICs are not connected!\n\n");
                print_status();
                return;
            }
        }

        // data and descriptors preparation

        uint32_t addr;
        int i;
        // clear transmitter buffer
//        addr = MCB_DPRAM_BASE(0) + DPRAM_HALF + 4;
//        for	(i = 0; i < 32767; i++)	{
//            mcb_write_reg(addr, 0);
//            addr += 4;
//        }

        // clear receiver buffer
        addr = MCB_DPRAM_BASE(0);
        for	(i = 0; i < NB_WORDS + 2; i++)	{
            mcb_write_reg(addr, 0);
            addr += 4;
        }

        // create transmitter data
        addr = MCB_DPRAM_BASE(0) + DPRAM_HALF + 8;
        for	(i = 0; i < NB_WORDS; i++)	{
            mcb_write_reg(addr, 0x55aa0000 + (i & 0xFFFF));
            addr += 4;
        }

        mcb_write_reg(MCB_DPRAM_BASE(0) + DPRAM_HALF, 0xA0000000 | NB_WORDS * 4);

        // packet transfer
        mcb_write_reg(MCB_SWIC_RX_DESC_IR(rx_chan), MCB_DPRAM_BASE(0));
        mcb_write_reg(MCB_SWIC_RX_DATA_IR(rx_chan), MCB_DPRAM_BASE(0) + 8);
        mcb_write_reg(MCB_SWIC_TX_DESC_IR(tx_chan), MCB_DPRAM_BASE(0) + DPRAM_HALF);
        mcb_write_reg(MCB_SWIC_TX_DATA_IR(tx_chan), MCB_DPRAM_BASE(0) + DPRAM_HALF + 8);

        mcb_write_reg(MCB_SWIC_RX_DESC_CSR(rx_chan), MC_DMA_CSR_WCX(0));
        mcb_write_reg(MCB_SWIC_RX_DATA_CSR(rx_chan), MC_DMA_CSR_WCX(NB_WORDS / 2 - 1));
        mcb_write_reg(MCB_SWIC_TX_DESC_CSR(tx_chan), MC_DMA_CSR_WCX(0));
        mcb_write_reg(MCB_SWIC_TX_DATA_CSR(tx_chan), MC_DMA_CSR_WCX(NB_WORDS / 2 - 1));

        mcb_write_reg(MCB_SWIC_RX_DATA_RUN(rx_chan), 1);
        mcb_write_reg(MCB_SWIC_RX_DESC_RUN(rx_chan), 1);
        mcb_write_reg(MCB_SWIC_TX_DATA_RUN(tx_chan), 1);
        mcb_write_reg(MCB_SWIC_TX_DESC_RUN(tx_chan), 1);

        timeout = 10000;
        while (timeout) {
            Sleep(10);
            timeout -= 10;

            qstr = mcb_read_reg(MCB_PCI_QSTR_PCI);
            if (qstr & qstrErrMask) {
                printf("ERROR QSTR VALUE:   %08X\n", qstr);
                break;
            }

            if (qstr & dmaFinMask)
                break;
        }

        uint32_t desc = mcb_read_reg(MCB_DPRAM_BASE(0));
        printf("descriptor     = %08X\n", desc);
        printf("descriptor src = %08X\n", mcb_read_reg(MCB_DPRAM_BASE(0) + DPRAM_HALF));
        printf("data 0         = %08X\n", mcb_read_reg(MCB_DPRAM_BASE(0) + 8));
        printf("data 0 src     = %08X\n", mcb_read_reg(MCB_DPRAM_BASE(0) + DPRAM_HALF + 8));

        if (desc == (0xA0000000 | NB_WORDS * 4))
            printf("\n--> Good rx descriptor!\n\n");
        else
            printf("\n--> Error in rx descriptor!\n\n");

        unsigned nErrorsCur = 0;
        addr = MCB_DPRAM_BASE(0) + 8;
        for (i = 0; i < NB_WORDS; ++i) {
            uint32_t value = mcb_read_reg(addr);
            if (value != (0x55aa0000u + (i & 0xFFFF))) {
                printf("data error: i = %d   value = %x\n", i, value);
                nErrors++;
                nErrorsCur++;
                break;
            }
            addr += 4;
        }

        if (nErrorsCur == 0)
            printf("--> Data is successfully transmitted!\n\n");
        else
            printf("--> There are errors in rx data!\n\n");

        print_status();

        printf("\nstep = %d\tnErrors = %d\n", step+1, nErrors);

        Sleep(100);
    }
}

int main(int argc, char **argv)
{
    printf("\n=== MCB-03PEM-PCI SWIC TEST ===\n\n");

    parse_args(argc, argv);

    if (!libpe_init()) {
        perror("Failed to initialize libmcb");
        exit(1);
    }

    printf("VENDOR ID/DEVICE ID: %08X\n",   mcb_read_reg(MCB_PCI_DEVICE_VENDOR_ID));
    printf("SWIC HW VERSION:     %08X\n\n", mcb_read_reg(MCB_SWIC_HW_VER(0)));

    test();

    mcb_close_and_exit();
    return 0;
}





